from gettext import gettext as _
from typing import List, Union, Iterable, Dict, Optional, Callable, \
    TYPE_CHECKING, Tuple, Iterator, Mapping, Sequence

from gi.repository import GLib

from blueman.plugins.AppletPlugin import AppletPlugin

if TYPE_CHECKING:
    from typing_extensions import TypedDict

    class SubmenuItemDict(TypedDict):
        text: str
        markup: bool
        icon_name: str
        sensitive: bool
        tooltip: Optional[str]
        callback: Callable[[], None]

    class MenuItemDictBase(SubmenuItemDict):
        id: int

    class MenuItemDict(MenuItemDictBase, total=False):
        submenu: Iterable["SubmenuItemDict"]


class MenuItem:
    def __init__(self, menu_plugin: "Menu", owner: AppletPlugin, priority: Tuple[int, int], text: Optional[str],
                 markup: bool, icon_name: Optional[str], tooltip: Optional[str], callback: Optional[Callable[[], None]],
                 submenu_function: Optional[Callable[[], Iterable["SubmenuItemDict"]]], visible: bool, sensitive: bool):
        self._menu_plugin = menu_plugin
        self._owner = owner
        self._priority = priority
        self._text = text
        self._markup = markup
        self._icon_name = icon_name
        self._tooltip = tooltip
        self._callback = callback
        self._submenu_function = submenu_function
        self._visible = visible
        self._sensitive = sensitive

        assert text and icon_name and (callback or submenu_function) or \
            not any([text, icon_name, tooltip, callback, submenu_function])

    @property
    def owner(self) -> AppletPlugin:
        return self._owner

    @property
    def priority(self) -> Tuple[int, int]:
        return self._priority

    @property
    def callback(self) -> Optional[Callable[[], None]]:
        return self._callback

    @property
    def visible(self) -> bool:
        return self._visible

    def _iter_base(self) -> Iterator[Tuple[str, Union[str, bool]]]:
        for key in ['text', 'markup', 'icon_name', 'tooltip', 'sensitive']:
            value = getattr(self, '_' + key)
            if value is not None:
                yield key, value

    def __iter__(self) -> Iterator[Tuple[str, Union[int, str, bool, List[Dict[str, Union[str, bool]]]]]]:
        yield "id", (self._priority[0] << 8) + self._priority[1]
        yield from self._iter_base()
        submenu = self.submenu_items
        if submenu:
            yield 'submenu', [dict(item) for item in submenu]

    @property
    def submenu_items(self) -> List["SubmenuItem"]:
        if not self._submenu_function:
            return []
        submenu_items = self._submenu_function()
        if not submenu_items:
            return []
        return [SubmenuItem(self._menu_plugin, self._owner, (0, 0), item.get('text'), item.get('markup', False),
                            item.get('icon_name'), item.get('tooltip'), item.get('callback'), None, True,
                            item.get('sensitive', True))
                for item in submenu_items]

    def set_text(self, text: str, markup: bool = False) -> None:
        self._text = text
        self._markup = markup
        self._menu_plugin.on_menu_changed()

    def set_icon_name(self, icon_name: str) -> None:
        self._icon_name = icon_name
        self._menu_plugin.on_menu_changed()

    def set_tooltip(self, tooltip: str) -> None:
        self._tooltip = tooltip
        self._menu_plugin.on_menu_changed()

    def set_visible(self, visible: bool) -> None:
        self._visible = visible
        self._menu_plugin.on_menu_changed()

    def set_sensitive(self, sensitive: bool) -> None:
        self._sensitive = sensitive
        self._menu_plugin.on_menu_changed()


class SubmenuItem(MenuItem):
    def __iter__(self) -> Iterator[Tuple[str, Union[str, bool]]]:
        yield from self._iter_base()


class Menu(AppletPlugin):
    __description__ = _("Provides a menu for the applet and an API for other plugins to manipulate it")
    __icon__ = "open-menu-symbolic"
    __author__ = "Walmis"
    __unloadable__ = False

    def on_load(self) -> None:
        self.__menuitems: Dict[Tuple[int, int], MenuItem] = {}

        self._add_dbus_signal("MenuChanged", "aa{sv}")
        self._add_dbus_method("GetMenu", (), "aa{sv}", self._get_menu)
        self._add_dbus_method("ActivateMenuItem", ("ai",), "", self._activate_menu_item)

    def add(self, owner: AppletPlugin, priority: Union[int, Tuple[int, int]], text: Optional[str] = None,
            markup: bool = False, icon_name: Optional[str] = None, tooltip: Optional[str] = None,
            callback: Optional[Callable[[], None]] = None,
            submenu_function: Optional[Callable[[], Iterable["SubmenuItemDict"]]] = None,
            visible: bool = True, sensitive: bool = True) -> MenuItem:

        if isinstance(priority, int):
            priority = (priority, 0)

        item = MenuItem(self, owner, priority, text, markup, icon_name, tooltip, callback, submenu_function, visible,
                        sensitive)
        self.__menuitems[item.priority] = item
        self.on_menu_changed()
        return item

    def unregister(self, owner: AppletPlugin) -> None:
        for item in list(self.__menuitems.values()):
            if item.owner == owner:
                del self.__menuitems[item.priority]
        self.on_menu_changed()

    def on_menu_changed(self) -> None:
        self._emit_dbus_signal("MenuChanged", self._get_menu())

    def _get_menu(self) -> List[Dict[str, GLib.Variant]]:
        return self._prepare_menu(dict(self.__menuitems[key])
                                  for key in sorted(self.__menuitems.keys())
                                  if self.__menuitems[key].visible)

    def _prepare_menu(self, data: Iterable[Mapping[str, Union[int, str, bool,
                                                              Iterable[Mapping[str, Union[str, bool]]]]]]) \
            -> List[Dict[str, GLib.Variant]]:
        return [{k: self._build_variant(v) for k, v in item.items()} for item in data]

    def _build_variant(self, value: Union[int, str, bool, Iterable[Mapping[str, Union[str, bool]]]]) -> GLib.Variant:
        if isinstance(value, bool):
            return GLib.Variant("b", value)
        if isinstance(value, int):
            return GLib.Variant("i", value)
        if isinstance(value, str):
            return GLib.Variant("s", value)
        return GLib.Variant("aa{sv}", self._prepare_menu(value))

    def _activate_menu_item(self, indexes: Sequence[int]) -> None:
        node = self.__menuitems[(indexes[0] >> 8, indexes[0] % (1 << 8))]
        for index in list(indexes)[1:]:
            node = node.submenu_items[index]
        if node.callback:
            node.callback()



from gettext import gettext as _
import os
import logging
import importlib
import signal
from typing import List, Optional

from blueman.gui.GenericList import GenericList, ListDataDict
import blueman.plugins.services
from blueman.plugins.ServicePlugin import ServicePlugin
from blueman.main.Config import Config

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GLib


class BluemanServices(Gtk.Application):
    def __init__(self) -> None:
        super().__init__(application_id="org.blueman.Services")
        self.window: Optional[Gtk.Window] = None

        def do_quit(_: object) -> bool:
            self.quit()
            return False

        s = GLib.unix_signal_source_new(signal.SIGINT)
        s.set_callback(do_quit)
        s.attach()

    def do_activate(self) -> None:
        if not self.window:
            self.window = Gtk.ApplicationWindow(application=self, title=_("Local Services"), icon_name="blueman",
                                                border_width=10)

            self.window.set_position(Gtk.WindowPosition.CENTER)

            grid = Gtk.Grid(orientation=Gtk.Orientation.VERTICAL, visible=True, row_spacing=10)
            self.window.add(grid)

            self.box = Gtk.Box(vexpand=True, visible=True)
            grid.add(self.box)

            button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, halign=Gtk.Align.END, visible=True)
            grid.add(button_box)

            self.b_apply = Gtk.Button(label=_("_Apply"), receives_default=True, use_underline=True,
                                      sensitive=False, visible=True, width_request=80)
            button_box.add(self.b_apply)

            self.viewport = Gtk.Viewport(visible=True, width_request=120)

            self.box.add(self.viewport)

            self.Config = Config("org.blueman.general")

            data: List[ListDataDict] = [
                {"id": "icon_name", "type": str, "renderer": Gtk.CellRendererPixbuf(stock_size=Gtk.IconSize.DND),
                 "render_attrs": {"icon_name": 0}},
                {"id": "caption", "type": str, "renderer": Gtk.CellRendererText(), "render_attrs": {"markup": 1},
                 "view_props": {"expand": True}},
                {"id": "id", "type": str},
            ]

            self.List = ls = GenericList(data, headers_visible=False, visible=True)

            ls.selection.connect("changed", self.on_selection_changed)

            self.viewport.add(ls)

            self.load_plugins()

            ls.selection.select_path(self.Config["services-last-item"])

            self.b_apply.connect("clicked", self.on_apply_clicked)

        self.window.present_with_time(Gtk.get_current_event_time())

    def option_changed(self) -> None:
        rets = [plugin.on_query_apply_state() for plugin in ServicePlugin.instances if plugin._is_loaded]
        show_apply = False
        for ret in rets:
            if ret == -1:
                show_apply = False
                break
            assert isinstance(ret, bool)
            show_apply = show_apply or ret

        self.b_apply.props.sensitive = show_apply

    def load_plugins(self) -> None:
        path = os.path.dirname(blueman.plugins.services.__file__)
        plugins = []
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith(".py") and not (f.endswith(".pyc") or f.endswith("_.py")):
                    plugins.append(f[0:-3])
        plugins.sort()
        logging.info(plugins)
        for plugin in plugins:
            try:
                importlib.import_module(f"blueman.plugins.services.{plugin}")
            except ImportError:
                logging.error(f"Unable to load {plugin} plugin", exc_info=True)

        for cls in ServicePlugin.__subclasses__():
            # FIXME this should not fail, if it does its a bug in the plugin
            try:
                inst = cls(self)
            except:  # noqa: E722
                logging.error(f"Failed to create instance of {cls}", exc_info=True)
                continue
            if not cls.__plugin_info__:
                logging.warning(f"Invalid plugin info in {cls}")
            else:
                (name, icon) = cls.__plugin_info__
                self.setup_list_item(inst, name, icon)

    def setup_list_item(self, inst: ServicePlugin, name: str, icon: str) -> None:
        self.List.append(icon_name=icon, caption=name, id=inst.__class__.__name__)

    def on_apply_clicked(self, _button: Gtk.Button) -> None:
        for plugin in ServicePlugin.instances:
            if plugin._is_loaded:
                plugin.on_apply()
        self.option_changed()

    def set_page(self, pageid: str) -> None:
        logging.info(f"Set page {pageid}")

        if len(ServicePlugin.instances) == 0:
            return
        # set the first item
        if pageid is None:
            pageid = ServicePlugin.instances[0].__class__.__name__
        for inst in ServicePlugin.instances:
            if inst.__class__.__name__ == pageid:
                if not inst._is_loaded:
                    inst.on_load(self.box)
                    inst._is_loaded = True

                inst._on_enter()
            else:
                inst._on_leave()

    def on_selection_changed(self, _selection: Gtk.TreeSelection) -> None:
        tree_iter = self.List.selected()
        assert tree_iter
        if self.List.get_cursor()[0]:
            # GtkTreePath returns row when used as string
            self.Config["services-last-item"] = int(str(self.List.get_cursor()[0]))
        row = self.List.get(tree_iter, "id")
        rowid = row["id"]

        self.set_page(rowid)



from gettext import gettext as _
from operator import itemgetter
import time
import logging
from typing import List, TYPE_CHECKING, Optional, Callable, cast, Union

from blueman.bluez.Device import Device
from blueman.bluez.errors import DBusNoSuchAdapterError
from blueman.gui.Notification import Notification
from blueman.Sdp import ServiceUUID
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.plugins.applet.PowerManager import PowerManager, PowerStateListener

if TYPE_CHECKING:
    from blueman.plugins.applet.Menu import SubmenuItemDict

    from typing_extensions import TypedDict

    class _ItemBase(TypedDict):
        adapter: str
        address: str
        alias: str
        icon: str
        name: str
        uuid: str

    class Item(_ItemBase):
        time: float
        device: Optional[str]
        mitem: Optional[SubmenuItemDict]

    class StoredIcon(_ItemBase):
        time: str

REGISTRY_VERSION = 0


class RecentConns(AppletPlugin, PowerStateListener):
    __depends__ = ["DBusService", "Menu"]
    __icon__ = "document-open-recent-symbolic"
    __description__ = _("Provides a menu item that contains last used connections for quick access")
    __author__ = "Walmis"

    __gsettings__ = {
        "schema": "org.blueman.plugins.recentconns",
        "path": None
    }
    __options__ = {
        "max-items": {"type": int,
                      "default": 6,
                      # the maximum number of items RecentConns menu will display
                      "name": _("Maximum items"),
                      "desc": _("The maximum number of items recent connections menu will display."),
                      "range": (6, 20)},
        "recent-connections": {"type": list, "default": "[]"}
    }

    _items = None

    def on_load(self) -> None:
        self.__menuitems: List["SubmenuItemDict"] = []

        self._item = self.parent.Plugins.Menu.add(self, 52, text=_("Recent _Connections") + "…",
                                                  icon_name="document-open-recent-symbolic",
                                                  submenu_function=self.get_menu)
        self.parent.Plugins.Menu.add(self, 53)

    def on_power_state_changed(self, manager: PowerManager, state: bool) -> None:
        self._rebuild()

    def on_unload(self) -> None:
        self.parent.Plugins.Menu.unregister(self)

    def _rebuild(self) -> None:
        if 'PowerManager' in self.parent.Plugins.get_loaded() and \
                not self.parent.Plugins.PowerManager.get_bluetooth_status():
            self._item.set_sensitive(False)
            return

        self._items = self._get_items()

        if len(self._items) == 0:
            self._item.set_sensitive(False)
            return

        self._item.set_sensitive(True)

        self.__menuitems = [self._build_menu_item(item) for item in self._items[:self.get_option("max-items")]]

        self.parent.Plugins.Menu.on_menu_changed()

    def on_manager_state_changed(self, state: bool) -> None:
        if state:
            self._rebuild()
        else:
            self._item.set_sensitive(False)

    def on_device_created(self, path: str) -> None:
        self._rebuild()

    def on_device_removed(self, path: str) -> None:
        self._rebuild()

    def on_adapter_added(self, path: str) -> None:
        self._rebuild()

    def on_adapter_removed(self, path: str) -> None:
        self._rebuild()

    def notify(self, object_path: str, uuid: str) -> None:
        device = Device(obj_path=object_path)
        logging.info(f"{device} {uuid}")
        try:
            adapter = self.parent.Manager.get_adapter(device['Adapter'])
        except DBusNoSuchAdapterError:
            logging.warning("adapter not found")
            return

        item = {
            "adapter": adapter["Address"],
            "address": device['Address'],
            "alias": device['Alias'],
            "icon": device['Icon'],
            "name": ServiceUUID(uuid).name,
            "uuid": uuid,
            "time": str(time.time()),
        }

        stored_items = self.get_option("recent-connections")

        for i in stored_items:
            if i["adapter"] == item["adapter"] and \
                    i["address"] == item["address"] and \
                    i["uuid"] == item["uuid"]:
                i["time"] = item["time"]
                i["device"] = object_path
                break
        else:
            stored_items.append(item)

        self.set_option("recent-connections", stored_items)

        self._rebuild()

    def on_item_activated(self, item: "Item") -> None:
        logging.info(f"Connect {item['address']} {item['uuid']}")

        assert item["mitem"] is not None

        item["mitem"]["sensitive"] = False
        self.parent.Plugins.Menu.on_menu_changed()

        def reply() -> None:
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            Notification(_("Connected"), _("Connected to %s") % item["mitem"]["text"],
                         icon_name=item["icon"]).show()
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        def err(reason: Union[Exception, str]) -> None:
            Notification(_("Failed to connect"), str(reason).split(": ")[-1],
                         icon_name="dialog-error").show()
            assert item["mitem"] is not None  # https://github.com/python/mypy/issues/2608
            item["mitem"]["sensitive"] = True
            self.parent.Plugins.Menu.on_menu_changed()

        self.parent.Plugins.DBusService.connect_service(item["device"], item["uuid"], reply, err)

    def _build_menu_item(self, item: "Item") -> "SubmenuItemDict":
        mitem: "SubmenuItemDict" = {
            "text": _("%(service)s on %(device)s") % {"service": item["name"], "device": item["alias"]},
            "markup": True,
            "icon_name": item["mitem"]["icon_name"] if item["mitem"] is not None else item["icon"],
            "sensitive": item["device"] is not None,
            "tooltip": None if item["device"] is None else _("Adapter for this connection is not available"),
            "callback": (item["mitem"]["callback"] if item["mitem"] is not None
                         else cast(Callable[[], None], lambda itm=item: self.on_item_activated(itm)))
        }

        item["mitem"] = mitem

        return mitem

    def get_menu(self) -> List["SubmenuItemDict"]:
        return self.__menuitems

    def _get_device_path(self, adapter_path: str, address: str) -> Optional[str]:
        try:
            adapter = self.parent.Manager.get_adapter(adapter_path)
        except DBusNoSuchAdapterError:
            return None

        device = self.parent.Manager.find_device(address, adapter.get_object_path())
        return device.get_object_path() if device is not None else None

    def _get_items(self) -> List["Item"]:
        adapter_addresses = {adapter["Address"] for adapter in self.parent.Manager.get_adapters()}

        return sorted(
            ({
                "adapter": i["adapter"],
                "address": i["address"],
                "alias": i["alias"],
                "icon": i["icon"],
                "name": i["name"],
                "uuid": i["uuid"],
                "time": float(i["time"]),
                "device": (self._get_device_path(i["adapter"], i["address"])
                           if i["adapter"] in adapter_addresses else None),
                "mitem": None
            }
                for i in self.get_option("recent-connections")
                if i["adapter"] not in adapter_addresses or self._get_device_path(i["adapter"], i["address"])),
            key=itemgetter("time"),
            reverse=True
        )



from collections import OrderedDict
from typing import Iterable, TYPE_CHECKING, Callable, List, Tuple, Dict, Union, TypeVar, Any

from gi.repository import Gio, GLib, Pango

from blueman.main.DbusService import DbusService
from blueman.main.Tray import BluemanTray
from blueman.main.indicators.IndicatorInterface import IndicatorInterface, IndicatorNotAvailable

if TYPE_CHECKING:
    from blueman.plugins.applet.Menu import MenuItemDict, SubmenuItemDict
    from blueman.main.indicators.GtkStatusIcon import MenuItemActivator


class MenuService(DbusService):
    def __init__(self, on_activate_menu_item: "MenuItemActivator") -> None:
        super().__init__(None, "com.canonical.dbusmenu", "/org/blueman/sni/menu", Gio.BusType.SESSION)
        self._items: OrderedDict[int, "MenuItemDict"] = OrderedDict()
        self._revision = 0
        self._revision_advertised = -1
        self._on_activate = on_activate_menu_item

        self.add_method("GetLayout", ("i", "i", "as"), ("u", "(ia{sv}av)"), self._get_layout)
        self.add_method("Event", ("i", "s", "v", "u"), (), self._on_event)
        self.add_method("AboutToShow", ("i",), ("b",), lambda _: self._revision > self._revision_advertised)

        self.add_method("GetGroupProperties", ("ai", "as"), ("a(ia{sv})",),
                        lambda ids, props: [(idx, self._render_item(item)) for idx, item in self._iterate_items()
                                            if idx in ids])

        self.add_signal("LayoutUpdated", ("u", "i"))

        GLib.timeout_add(100, self._advertise_revision)

    def set_items(self, items: Iterable["MenuItemDict"]) -> None:
        self._items = OrderedDict((item["id"], item) for item in items)
        self._revision += 1

    def _advertise_revision(self) -> bool:
        if self._revision != self._revision_advertised:
            self.emit_signal("LayoutUpdated", self._revision, 0)
            self._revision_advertised = self._revision
        return True

    def _get_layout(self, parent_id: int, _recursion_depth: int, _property_names: List[str]
                    ) -> Tuple[int, Tuple[int, Dict[str, GLib.Variant], List[GLib.Variant]]]:
        if parent_id == 0:
            return self._revision, (0, {}, self._render_menu(((item["id"] << 8, item) for item in self._items.values()),
                                                             self._render_submenu))
        else:
            item = self._items[parent_id >> 8]
            if "submenu" in item and _recursion_depth != 0:
                return self._revision, (parent_id, self._render_item(item), self._render_submenu(item, parent_id))
            return self._revision, (parent_id, self._render_item(item), [])

    def _render_submenu(self, item: "MenuItemDict", idx: int) -> List[GLib.Variant]:
        if "submenu" in item:
            return self._render_menu(enumerate(item["submenu"], idx + 1), lambda _item, _isx: [])
        else:
            return []

    _T = TypeVar("_T", bound="SubmenuItemDict")

    def _render_menu(self, items: Iterable[Tuple[int, _T]], submenu_callback: Callable[[_T, int], List[GLib.Variant]]
                     ) -> List[GLib.Variant]:
        return [GLib.Variant("(ia{sv}av)", (idx, self._render_item(item), submenu_callback(item, idx)))
                for (idx, item) in items]

    def _iterate_items(self) -> Iterable[Tuple[int, "SubmenuItemDict"]]:
        for item in self._items.values():
            yield item["id"] << 8, item
            if "submenu" in item:
                yield from enumerate(item["submenu"], (item["id"] << 8) + 1)

    def _render_item(self, item: Union["MenuItemDict", "SubmenuItemDict"]) -> Dict[str, GLib.Variant]:
        if "text" in item and "icon_name" in item:
            label = Pango.parse_markup(item["text"], -1, "\0")[2] if item.get("markup", False) else item["text"]
            props = {
                "label": GLib.Variant("s", label),
                "icon-name": GLib.Variant("s", item["icon_name"]),
                "enabled": GLib.Variant("b", item["sensitive"]),
            }
            if "submenu" in item:
                props["children-display"] = GLib.Variant("s", "submenu")
            return props
        else:
            return {"type": GLib.Variant("s", "separator")}

    def _on_event(self, idx: int, event_id: str, _data: GLib.Variant, _timestamp: int) -> None:
        if event_id == "clicked":
            if idx % (1 << 8) == 0:
                self._on_activate(idx >> 8)
            else:
                self._on_activate(idx >> 8, idx % (1 << 8) - 1)


class StatusNotifierItemService(DbusService):
    Category = "Hardware"
    Id = "blueman"
    Title = "blueman"
    ItemIsMenu = False

    def __init__(self, tray: BluemanTray, icon_name: str) -> None:
        super().__init__(None, "org.kde.StatusNotifierItem", "/org/blueman/sni", Gio.BusType.SESSION,
                         {"Category": "s", "Id": "s", "IconName": "s", "Status": "s", "Title": "s",
                          "ToolTip": "(sa(iiay)ss)", "Menu": "o", "ItemIsMenu": "b"})
        self.add_method("Activate", ("i", "i"), "", lambda x, y: tray.activate_status_icon())

        self.menu = MenuService(tray.activate_menu_item)

        self.IconName = icon_name
        self.Status = "Active"
        self.ToolTip: Tuple[str, List[Tuple[int, int, List[int]]], str, str] = ("", [], "", "")
        self.Menu = "/org/blueman/sni/menu"

        self.add_signal("NewIcon", "")
        self.add_signal("NewStatus", "s")
        self.add_signal("NewToolTip", "")

    def register(self) -> None:
        self.menu.register()
        super().register()

    def unregister(self) -> None:
        super().unregister()
        self.menu.unregister()


class StatusNotifierItem(IndicatorInterface):
    _SNI_BUS_NAME = _SNI_INTERFACE_NAME = "org.kde.StatusNotifierWatcher"

    def __init__(self, tray: BluemanTray, icon_name: str) -> None:
        self._sni = StatusNotifierItemService(tray, icon_name)
        self._sni.register()

        self._bus = Gio.bus_get_sync(Gio.BusType.SESSION)

        watcher_expected: bool

        def on_watcher_appeared(*args: Any) -> None:
            nonlocal watcher_expected

            if watcher_expected:
                watcher_expected = False
            else:
                tray.activate()

        Gio.bus_watch_name(Gio.BusType.SESSION, self._SNI_BUS_NAME, Gio.BusNameWatcherFlags.NONE,
                           on_watcher_appeared, None)

        try:
            Gio.bus_get_sync(Gio.BusType.SESSION).call_sync(
                self._SNI_BUS_NAME, "/StatusNotifierWatcher", self._SNI_INTERFACE_NAME,
                "RegisterStatusNotifierItem", GLib.Variant("(s)", ("/org/blueman/sni",)),
                None, Gio.DBusCallFlags.NONE, -1)
            watcher_expected = True
        except GLib.Error:
            watcher_expected = False
            raise IndicatorNotAvailable

    def set_icon(self, icon_name: str) -> None:
        self._sni.IconName = icon_name
        self._sni.emit_signal("NewIcon")

    def set_tooltip_title(self, title: str) -> None:
        self._sni.ToolTip = ("", [], title, self._sni.ToolTip[3])
        self._sni.emit_signal("NewToolTip")

    def set_tooltip_text(self, text: str) -> None:
        self._sni.ToolTip = ("", [], self._sni.ToolTip[2], text)
        self._sni.emit_signal("NewToolTip")

    def set_visibility(self, visible: bool) -> None:
        self._sni.Status = status = "Active" if visible else "Passive"
        self._sni.emit_signal("NewStatus", status)

    def set_menu(self, menu: Iterable["MenuItemDict"]) -> None:
        self._sni.menu.set_items(menu)



__all__ = ["VERSION", "PACKAGE", "WEBSITE", "ICON_PATH", "PIXMAP_PATH", "UI_PATH", "BIN_DIR"]

VERSION = "2.3.5"
PACKAGE = "blueman"
WEBSITE = "https://github.com/blueman-project/blueman"
PREFIX = "/usr"
BIN_DIR = "/usr/bin"
LOCALEDIR = "/usr/share/locale"
ICON_PATH = "/usr/share/icons"
PIXMAP_PATH = "/usr/share/blueman/pixmaps"
UI_PATH = "/usr/share/blueman/ui"
DHCP_CONFIG_FILE = "/etc/dhcpd.conf"
POLKIT = False
GETTEXT_PACKAGE = "blueman"
RFCOMM_WATCHER_PATH = "/usr/lib/blueman/blueman-rfcomm-watcher"

import os

if 'BLUEMAN_SOURCE' in os.environ:
    _dirname = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    BIN_DIR = os.path.join(_dirname, 'apps')
    ICON_PATH = os.path.join(_dirname, 'data', 'icons')
    PIXMAP_PATH = os.path.join(_dirname, 'data', 'icons', 'pixmaps')
    UI_PATH = os.path.join(_dirname, 'data', 'ui')



from gettext import gettext as _
from typing import Optional, Tuple, List

from gi.repository import GObject, GLib, Gio

from blueman.Functions import launch
from blueman.main.Config import Config
from blueman.main.PluginManager import PluginManager
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluemantyping import GSignals


class StatusIconImplementationProvider:
    def on_query_status_icon_implementation(self) -> Tuple[str, int]:
        ...


class StatusIconVisibilityHandler:
    def on_query_force_status_icon_visibility(self) -> bool:
        ...


class StatusIconProvider:
    def on_status_icon_query_icon(self) -> Optional[str]:
        ...


class StatusIcon(AppletPlugin, GObject.GObject):
    __gsignals__: GSignals = {'activate': (GObject.SignalFlags.NO_HOOKS, None, ())}

    __unloadable__ = False
    __icon__ = "bluetooth-symbolic"
    __depends__ = ['Menu']

    visible = None

    visibility_timeout = None

    _implementations = None

    def on_load(self) -> None:
        GObject.GObject.__init__(self)
        self._tooltip_title = _("Bluetooth Enabled")
        self._tooltip_text = ""

        self.general_config = Config("org.blueman.general")
        self.general_config.connect("changed::symbolic-status-icons", self.on_symbolic_config_change)

        self.query_visibility(emit=False)

        self.parent.Plugins.connect('plugin-loaded', self._on_plugins_changed)
        self.parent.Plugins.connect('plugin-unloaded', self._on_plugins_changed)

        self._add_dbus_method("GetVisibility", (), "b", lambda: self.visible)
        self._add_dbus_signal("VisibilityChanged", "b")
        self._add_dbus_signal("ToolTipTitleChanged", "s")
        self._add_dbus_signal("ToolTipTextChanged", "s")
        self._add_dbus_method("GetToolTipTitle", (), "s", lambda: self._tooltip_title)
        self._add_dbus_method("GetToolTipText", (), "s", lambda: self._tooltip_text)
        self._add_dbus_signal("IconNameChanged", "s")
        self._add_dbus_method("GetStatusIconImplementations", (), "as", self._get_status_icon_implementations)
        self._add_dbus_method("GetIconName", (), "s", self._get_icon_name)
        self._add_dbus_method("Activate", (), "", lambda: self.emit("activate"))

    def query_visibility(self, delay_hiding: bool = False, emit: bool = True) -> None:
        if self.parent.Manager.get_adapters() or \
           any(plugin.on_query_force_status_icon_visibility()
               for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconVisibilityHandler)):
            self.set_visible(True, emit)
        elif not self.visibility_timeout:
            if delay_hiding:
                self.visibility_timeout = GLib.timeout_add(2500, self.on_visibility_timeout)
            else:
                self.set_visible(False, emit)

    def on_visibility_timeout(self) -> bool:
        assert self.visibility_timeout is not None
        GLib.source_remove(self.visibility_timeout)
        self.visibility_timeout = None
        self.query_visibility()
        return False

    def set_visible(self, visible: bool, emit: bool) -> None:
        self.visible = visible
        if emit:
            self._emit_dbus_signal("VisibilityChanged", visible)

    def set_tooltip_title(self, title: str) -> None:
        self._tooltip_title = title
        self._emit_dbus_signal("ToolTipTitleChanged", title)

    def set_tooltip_text(self, text: Optional[str]) -> None:
        self._tooltip_text = "" if text is None else text
        self._emit_dbus_signal("ToolTipTextChanged", self._tooltip_text)

    def on_symbolic_config_change(self, settings: Gio.Settings, key: str) -> None:
        self.icon_should_change()

    def icon_should_change(self) -> None:
        self._emit_dbus_signal("IconNameChanged", self._get_icon_name())
        self.query_visibility()

    def on_adapter_added(self, _path: str) -> None:
        self.query_visibility()

    def on_adapter_removed(self, _path: str) -> None:
        self.query_visibility()

    def on_manager_state_changed(self, state: bool) -> None:
        self.query_visibility()
        if state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _on_plugins_changed(self, _plugins: PluginManager, _name: str) -> None:
        implementations = self._get_status_icon_implementations()
        if not self._implementations or self._implementations != implementations:
            self._implementations = implementations

        if self.parent.manager_state:
            launch('blueman-tray', icon_name='blueman', sn=False)

    def _get_status_icon_implementations(self) -> List[str]:
        return [implementation for implementation, _ in sorted(
            (plugin.on_query_status_icon_implementation()
             for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconImplementationProvider)),
            key=lambda implementation_priority: implementation_priority[1],
            reverse=True
        )] + ["GtkStatusIcon"]

    def _get_icon_name(self) -> str:
        # default icon name
        name = "blueman-tray"
        for plugin in self.parent.Plugins.get_loaded_plugins(StatusIconProvider):
            icon = plugin.on_status_icon_query_icon()
            if icon is not None:
                # status icon
                name = icon

        # depending on configuration, ensure fullcolor icons..
        name = name.replace("-symbolic", "")
        if self.general_config.get_boolean("symbolic-status-icons"):
            # or symbolic
            name = f"{name}-symbolic"

        return name
